/**
 * Copyright (C) 2012  Dušan Vejnovič  <vaadin@dussan.org>
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.dussan.vaadin.dtable;

import java.text.DateFormat;
import java.text.DecimalFormat;
import java.text.NumberFormat;
import java.text.SimpleDateFormat;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collections;
import java.util.Date;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Locale;
import java.util.Map;

import org.dussan.vaadin.dtable.client.ui.VDTable;
import org.dussan.vaadin.dtable.helpers.ManifestHelper;

import com.vaadin.terminal.PaintException;
import com.vaadin.terminal.PaintTarget;
import com.vaadin.terminal.gwt.server.WebApplicationContext;
import com.vaadin.terminal.gwt.server.WebBrowser;
import com.vaadin.ui.AbstractComponent;
import com.vaadin.ui.ClientWidget;
import com.vaadin.ui.Window.Notification;

@SuppressWarnings("serial")
@ClientWidget(VDTable.class)
public abstract class DTable extends AbstractComponent {

	private static final int DEFAULT_TABLE_HEIGHT = 300;
	private static final int DEFAULT_TABLE_FROZEN_MINIMAL_WIDTH = 150;
	private static final int DEFAULT_TABLE_MINIMAL_WIDTH = 300;
	private static final int DEFAULT_HEADER_ROW_HEIGHT = 26;
	private static final int DEFAULT_BODY_ROW_HEIGHT = 22;
	private static final int DEFAULT_PAGER_PAGES = 15;
	private static final int DEFAULT_PAGE_SIZE = 15;
	private static final int DEFAULT_COLUMN_WIDTH = 50;
	private static final HorizontalAlignment DEFAULT_HEADER_ALIGNMENT = HorizontalAlignment.CENTER;
	private static final HorizontalAlignment DEFAULT_COLUMN_ALIGNMENT = HorizontalAlignment.LEFT;
	private static final int DEFAULT_SCREEN_WIDTH = 4000;

	private boolean isInitialized = false;
	private boolean enableAllColumns = false;
	private boolean enableAllRows = false;
	private boolean enableFrozenTable = false;
	private int left = 0;
	private int top = 0;
	private int firstColumn = 0;
	private int firstRow = 0;
	private int allColumns = 0;
	private long allRows = 0;
	private int tableWidth = DEFAULT_SCREEN_WIDTH;
	private int columnDefaultWidth = DEFAULT_COLUMN_WIDTH;
	private int tableHeight = DEFAULT_TABLE_HEIGHT;
	private int tableFrozenMinimalWidth = DEFAULT_TABLE_FROZEN_MINIMAL_WIDTH;
	private int tableMinimalWidth = DEFAULT_TABLE_MINIMAL_WIDTH;
	private int headerRowHeight = DEFAULT_HEADER_ROW_HEIGHT;
	private int bodyRowHeight = DEFAULT_BODY_ROW_HEIGHT;
	private int pagerPages = DEFAULT_PAGER_PAGES;
	private int pageSize = DEFAULT_PAGE_SIZE;
	private String pageReloadContent = new String("false:0:0:0:0:0:0");
	private HorizontalAlignment headerDefaultAlignment = DEFAULT_HEADER_ALIGNMENT;
	private HorizontalAlignment columnDefaultAlignment = DEFAULT_COLUMN_ALIGNMENT;
	private String noDataMessageTitle = "NO NEW DATA";
	private String noDataMessage = "Size of the requested data is zero!";
	private String errorMessageTitle = "ERROR";
	private String errorMessage = "When transmitting data an error has occurred!";

	private Locale formattingLocale = null;
	private Map<String, Object> paintContentObjects = new HashMap<String, Object>();
	private Map<String, Object> paintContentObjectsArchive = new HashMap<String, Object>();
	private List<String> columnHeaders = new ArrayList<String>();
	private List<String> columnHeadersId = new ArrayList<String>();
	private Map<Integer, Integer> headerAlignments = new HashMap<Integer, Integer>();
	private Map<Integer, Integer> columnAlignments = new HashMap<Integer, Integer>();
	private Map<Integer, Integer> columnWidths = new HashMap<Integer, Integer>();
	private Map<Integer, Object[]> dateTimeFormats = new HashMap<Integer, Object[]>();
	private Map<Integer, String> numberFormats = new HashMap<Integer, String>();
	private Map<Integer, Filter> columnFilters = new HashMap<Integer, Filter>();
	private List<Integer> columnFormatted = new ArrayList<Integer>();
	private List<Integer> columnHiddens = new ArrayList<Integer>();
	private List<Integer> columnFrozens = new ArrayList<Integer>();
	private List<Integer> columnVisibles = new ArrayList<Integer>();
	private List<String> columnSortings = new ArrayList<String>();
	private List<Integer> columnFreezeEnabled = new ArrayList<Integer>();
	private List<Integer> columnUnfreezeEnabled = new ArrayList<Integer>();
	private List<List<Object>> columnsData = new ArrayList<List<Object>>();

	public DTable() {
		// don't remove: it define dynamic widget's width
		setWidth("100%");
	}

	public static String getVersion() {
		if (ManifestHelper.getManifest() != null) {
			return ManifestHelper.getManifest().getMainAttributes()
					.getValue("Implementation-Version");
		}
		return null;
	}

	public static String getGitVersion() {
		if (ManifestHelper.getManifest() != null) {
			return ManifestHelper.getManifest().getMainAttributes()
					.getValue("Git-Version");
		}
		return null;
	}

	public static String getGwtQueryVersion() {
		if (ManifestHelper.getManifest() != null) {
			return ManifestHelper.getManifest().getMainAttributes()
					.getValue("GwtQuery-Version");
		}
		return null;
	}

	private void initTable() {
		isInitialized = true;
		checkColumnHeadersId();
		checkColumnWidths();
		checkVisibleColumns();

		if (paintContentObjects.containsKey("enablePageableTable")) {
			boolean enable = Boolean.parseBoolean(paintContentObjects.get(
					"enablePageableTable").toString());
			pageSize = !enable ? (tableHeight / bodyRowHeight + 2) : pageSize;
		}

		paintContentObjects.put("row", firstRow);
		paintContentObjects.put("left", left);
		paintContentObjects.put("top", top);
		paintContentObjects.put("pageSize", pageSize);
		paintContentObjects.put("tableHeight", tableHeight);
		paintContentObjects.put("tableFrozenMinimalWidth",
				tableFrozenMinimalWidth);
		paintContentObjects.put("tableMinimalWidth", tableMinimalWidth);
		paintContentObjects.put("headerRowHeight", headerRowHeight);
		paintContentObjects.put("bodyRowHeight", bodyRowHeight);
		paintContentObjects.put("columnWidths", getColumnWidths());
		paintContentObjects.put("headerAligments", getHeaderAlignments());
		paintContentObjects.put("columnAlignments", getColumnAlignments());

		if (columnFrozens != null && !columnFrozens.isEmpty()) {
			paintContentObjects.put("frozenColumns", getFrozenColumns());
		}

		if (columnHiddens != null && !columnHiddens.isEmpty()) {
			paintContentObjects.put("hiddenColumns", getColumnHiddens());
		}

		if (columnSortings != null && !columnSortings.isEmpty()) {
			paintContentObjects.put("sortings", getColumnSortingStrings());
		}

		if (columnFormatted != null && !columnFormatted.isEmpty()) {
			paintContentObjects.put("formatedColumns", getColumnFormatted());
		}

		if (columnFreezeEnabled != null && !columnFreezeEnabled.isEmpty()) {
			paintContentObjects.put("enableFreezeColumns",
					columnFreezeEnabled.toArray(new Integer[] {}));
		}

		if (columnUnfreezeEnabled != null && !columnUnfreezeEnabled.isEmpty()) {
			paintContentObjects.put("enableUnfreezeColumns",
					columnUnfreezeEnabled.toArray(new Integer[] {}));
		}

		paintContentObjects.put("visibleColumns", getVisibleColumns());
		paintContentObjects.put("allRows", allRows = getAllRows(null));
		paintContentObjects.put("firstRow", 0);

		Map<Integer, String[]> data = getColumnsDatas(firstRow, pageSize,
				getFrozenColumns(), getVisibleColumns(), getColumnSortings(),
				null, true);
		if (data != null && !data.isEmpty()) {
			for (Map.Entry<Integer, String[]> entry : data.entrySet()) {
				paintContentObjects.put((isColumnFrozen(entry.getKey()) ? "ff"
						: "vv") + entry.getKey(), entry.getValue());
			}
		} else {
			paintContentObjects.put("emptyData", "emptyData");
		}
		paintContentObjectsArchive.putAll(paintContentObjects);
	}

	public void refreshLayout(int delayRefresh) {
		paintContentObjects.put("pageRefresh", delayRefresh <= 0 ? 500
				: delayRefresh);
	}

	@Override
	public void attach() {
		if (!isInitialized) {
			initTable();
		}
		requestRepaint();
	}

	@Override
	public void paintContent(PaintTarget target) throws PaintException {
		super.paintContent(target);

		if (!isInitialized) {
			initTable();
		} else {
			paintContentObjects.put("row", firstRow);
			paintContentObjects.put("left", left);
			paintContentObjects.put("top", top);
			paintContentObjects.put("frozenColumns", getFrozenColumns());
			paintContentObjects.put("visibleColumns", getVisibleColumns());
			paintContentObjects
					.put("allColumnsVisible",
							allColumns == (columnFrozens.size()
									+ columnHiddens.size() + columnVisibles
									.size()));
		}

		for (Map.Entry<String, Object> entry : paintContentObjects.entrySet()) {
			String name = entry.getKey();
			Object value = entry.getValue();
			if (value instanceof Boolean) {
				target.addAttribute(name, (Boolean) value);
			} else if (value instanceof Double) {
				target.addAttribute(name, (Double) value);
			} else if (value instanceof Float) {
				target.addAttribute(name, (Float) value);
			} else if (value instanceof Integer) {
				target.addAttribute(name, (Integer) value);
			} else if (value instanceof Long) {
				target.addAttribute(name, (Long) value);
			} else if (value instanceof String) {
				target.addAttribute(name, (String) value);
			} else if (value instanceof String[] || value instanceof Integer[]) {
				target.addAttribute(name, (Object[]) value);
			}
		}
		paintContentObjects.clear();

		if (paintContentObjectsArchive.containsKey("pageReload")) {
			paintContentObjectsArchive.remove("pageReload");
		}
	}

	@SuppressWarnings("unchecked")
	@Override
	public void changeVariables(Object source, Map<String, Object> variables) {
		super.changeVariables(source, variables);
		Map<String, Object> request = (Map<String, Object>) variables
				.get("request");
		if (request == null || request.isEmpty()
				|| request.containsKey("pageReload")) {
			paintContentObjects.putAll(paintContentObjectsArchive);
			paintContentObjects.put("pageReload", pageReloadContent);
			requestRepaint();
		} else if (!request.containsKey("top") || !request.containsKey("left")
				|| !request.containsKey("width")
				|| !request.containsKey("pageSize")
				|| !request.containsKey("prContent")) {
			// "NAPAKA", "Pri prenosu podatkov je prišlo do napake!"
			getApplication().getMainWindow().showNotification(
					errorMessageTitle, errorMessage,
					Notification.TYPE_ERROR_MESSAGE);
		} else if (request.containsKey("noData")) {
			// "NI NOVIH PODATKOV", "Velikost zahtevanih podatkov je nič!",
			getApplication().getMainWindow().showNotification(
					noDataMessageTitle, noDataMessage,
					Notification.TYPE_ERROR_MESSAGE);
		} else {
			Boolean fetchNewData = null;
			Map<Integer, Filter> filters = new HashMap<Integer, Filter>(
					columnFilters);

			// standard message elements
			left = Integer.parseInt(request.get("left").toString());
			top = Integer.parseInt(request.get("top").toString());
			tableWidth = Integer.parseInt(request.get("width").toString());
			pageSize = Integer.parseInt(request.get("pageSize").toString());
			pageReloadContent = request.get("prContent").toString();

			if (request.containsKey("newFilter")) {
				String[] newFilter = request.get("newFilter").toString()
						.split("_:_");

				if (newFilter != null && newFilter.length == 5) {
					Integer column = Integer.parseInt(newFilter[0]);
					Integer row = Integer.parseInt(newFilter[1]);
					String columnHeader = columnHeaders.get(column);
					String columnHeaderId = columnHeadersId.get(column);
					String newOperator = newFilter[2];
					String newDescription = newFilter[3];
					String newFilterValue = newFilter[4];
					String newValue = formatData(column, columnsData
							.get(column).get(row));
					String newOriginalValue = columnsData.get(column).get(row) == null ? new String()
							: columnsData.get(column).get(row).toString();

					int type = Filter.TEXT;
					String formatFrom = new String();
					String formatTo = new String();
					List<String> operators = new ArrayList<String>();
					List<String> values = new ArrayList<String>();
					List<String> originalValues = new ArrayList<String>();

					if (filters.containsKey(column)) {
						Filter filter = filters.get(column);
						type = filter.getType();
						formatFrom = filter.getFormatFrom();
						formatTo = filter.getFormatTo();
						operators.addAll(Arrays.asList(filter.getOperators()));
						values.addAll(Arrays.asList(filter.getValues()));
						originalValues.addAll(Arrays.asList(filter
								.getOriginalValues()));
					} else {
						if (numberFormats.containsKey(column)) {
							type = Filter.NUMBER;
							formatTo = numberFormats.get(column);
						} else if (dateTimeFormats.containsKey(column)) {
							Object[] dateTimeFormat = dateTimeFormats
									.get(column);
							if (!Boolean.parseBoolean(dateTimeFormat[2]
									.toString())) {
								type = Filter.DATE_TIME;
							} else {
								type = Filter.STANDARD_DATE_TIME;
							}
							formatFrom = dateTimeFormat[0].toString();
							formatTo = dateTimeFormat[1].toString();
						}
					}

					// check filter with 'LIKE' operator and
					// '%' symbol in description
					if (newOperator.equals(Filter.LIKE)
							&& newFilterValue != null
							&& newFilterValue.indexOf("%") >= 0) {
						int substrLength = newValue.length();
						if (newFilterValue.indexOf("%") != newFilterValue
								.lastIndexOf("%")) {
							substrLength = newFilterValue.lastIndexOf("%")
									- newFilterValue.indexOf("%") - 1;
						}
						newValue = newDescription
								.replace(Filter.LIKE + " ", "").replace("_",
										newValue.subSequence(0, substrLength));
						newValue = newValue.substring(1, newValue.length() - 1);
						newOriginalValue = newDescription.replace(
								Filter.LIKE + " ", "").replace("_",
								newOriginalValue.subSequence(0, substrLength));
						newOriginalValue = newOriginalValue.substring(1,
								newOriginalValue.length() - 1);
					}

					operators.add(newOperator);
					values.add(newValue);
					originalValues.add(newOriginalValue);
					filters.put(
							column,
							new Filter(column, columnHeader, columnHeaderId,
									type, formatFrom, formatTo, operators
											.toArray(new String[] {}), values
											.toArray(new String[] {}),
									originalValues.toArray(new String[] {})));

					paintContentObjects.put("newFilter",
							new String[] { column.toString(), newDescription,
									newFilterValue });
				}
			} else if (request.containsKey("removeFilter")) {
				Integer column = Integer.parseInt(request.get("removeFilter")
						.toString());
				if (filters.containsKey(column)) {
					filters.remove(column);
					paintContentObjects.put("removeFilter", column);
				} else {
					paintContentObjects.remove("removeFilter");
				}
			}

			if (request.containsKey("newData") || request.containsKey("freeze")
					|| request.containsKey("sortings")
					|| paintContentObjects.containsKey("newFilter")
					|| paintContentObjects.containsKey("removeFilter")) {
				Filter[] filterArray = new ArrayList<Filter>(filters.values())
						.toArray(new Filter[] {});

				if (request.containsKey("newData")) {
					if ("informServer".equals(request.get("newData"))) {
						fetchNewData = false;
					} else {
						fetchNewData = Boolean.parseBoolean(request.get(
								"newData").toString());
					}
				}

				if (request.containsKey("freeze") && enableFrozenTable) {
					int column = Integer.parseInt(request.get("freeze")
							.toString());
					if (isColumnFrozen(column)) {
						columnFrozens.remove(columnFrozens.indexOf(column));
					} else {
						setColumnFrozens(column);
					}
					fetchNewData = Boolean.TRUE;

					paintContentObjects
							.put("columnHeaders", getColumnHeaders());
					paintContentObjects.put("headerAligments",
							getHeaderAlignments());
					paintContentObjects.put("sortings",
							getColumnSortingStrings());
				}

				if (request.containsKey("sortings")) {
					fetchNewData = Boolean.TRUE;
					columnSortings.clear();
					if (request.get("sortings") != null
							&& !"+1-".equals(request.get("sortings"))) {
						columnSortings.addAll(Arrays.asList(request
								.get("sortings").toString().split("_:_")));
					}
				}

				if (paintContentObjects.containsKey("newFilter")
						|| paintContentObjects.containsKey("removeFilter")) {
					long allRows = getAllRows(filterArray);
					if (allRows == 0) {
						paintContentObjects.remove("newFilter");
						paintContentObjects.remove("removeFilter");
						paintContentObjects.put("noData", new String());
					} else {
						if (paintContentObjects.containsKey("newFilter")
								|| paintContentObjects
										.containsKey("removeFilter")) {
							fetchNewData = true;
							columnFilters = new HashMap<Integer, Filter>(
									filters);
							paintContentObjects.put("allRows",
									this.allRows = allRows);
						}
					}
				}

				calculateLeftAndTopValues();
				checkVisibleColumns();
				if (!variables.containsKey("noData") && fetchNewData != null) {
					// remove old column values
					for (String key : new HashSet<String>(
							paintContentObjectsArchive.keySet())) {
						if (key.length() > 0
								&& (key.startsWith("ff") || key
										.startsWith("vv"))) {
							paintContentObjectsArchive.remove(key);
						}
					}

					// remove old column values
					for (String key : new HashSet<String>(
							paintContentObjects.keySet())) {
						if (key.length() > 0
								&& (key.startsWith("ff") || key
										.startsWith("vv"))) {
							paintContentObjects.remove(key);
						}
					}

					paintContentObjects
							.put("frozenColumns", getFrozenColumns());
					paintContentObjects.put("visibleColumns",
							getVisibleColumns());

					// get new column values
					Map<Integer, String[]> data = getColumnsDatas(firstRow,
							pageSize, getFrozenColumns(), getVisibleColumns(),
							getColumnSortings(), filterArray, fetchNewData);
					if (data != null && !data.isEmpty()) {
						for (Map.Entry<Integer, String[]> entry : data
								.entrySet()) {
							paintContentObjects.put(
									(isColumnFrozen(entry.getKey()) ? "ff"
											: "vv") + entry.getKey(),
									entry.getValue());
						}
						paintContentObjects.put("allRowsVisible",
								allRows == data.size());
					} else {
						paintContentObjects.put("emptyData", "emptyData");
					}
				} else {
					paintContentObjects.put("emptyData", "emptyData");
				}

				paintContentObjectsArchive.putAll(paintContentObjects);
				if (!"informServer".equals(request.get("newData"))) {
					requestRepaint();
				}
			}
		}
	}

	abstract public long getAllRows(Filter[] filters);

	abstract public Object[] getRowsData(int startRow, int rows,
			Sorting[] sortings, Filter[] filters);

	private Map<Integer, String[]> getColumnsDatas(int startRow, int rows,
			Integer[] frozenColumns, Integer[] visibleColumns,
			Sorting[] sortings, Filter[] filters, boolean fetchNewData) {
		if (fetchNewData) {
			if (enableAllRows) {
				rows = 0;
				startRow = 0;
			}
			Object[] data = getRowsData(startRow, rows, sortings, filters);
			if (data != null && data.length > 0) {
				transformRows2ColumnsData(data);
			} else {
				return null;
			}
		}
		return getVisibleColumnsData(frozenColumns, visibleColumns);
	}

	private Map<Integer, String[]> getVisibleColumnsData(
			Integer[] frozenColumns, Integer[] visibleColumns) {
		List<Integer> columns = new ArrayList<Integer>();
		if (frozenColumns != null && frozenColumns.length > 0) {
			columns.addAll(Arrays.asList(frozenColumns));
		}
		if (visibleColumns != null && visibleColumns.length > 0) {
			columns.addAll(Arrays.asList(visibleColumns));
		}

		Map<Integer, String[]> tempColumnsData = new HashMap<Integer, String[]>();
		for (int column : columns) {
			if (column < allColumns) {
				tempColumnsData.put(column,
						prepareColumnsData(column, columnsData.get(column)));
			}
		}

		return tempColumnsData;
	}

	private String[] prepareColumnsData(int column, List<Object> columnsData) {
		List<String> data = new ArrayList<String>();
		for (int row = 0; row < columnsData.size(); row++) {
			data.add(formatData(column, columnsData.get(row)));
		}
		return data.toArray(new String[] {});
	}

	private void transformRows2ColumnsData(Object[] rowsData) {
		columnsData.clear();

		int allRows = allColumns == 0 ? 0 : rowsData.length / allColumns;
		for (int column = 0; column < allColumns; column++) {
			List<Object> columnData = new ArrayList<Object>();
			for (int row = 0; row < allRows; row++) {
				columnData.add(rowsData[allColumns * row + column]);
			}
			columnsData.add(columnData);
		}
	}

	private String formatData(int column, Object data) {
		String value = (data == null ? new String() : data.toString());

		// number format
		if (numberFormats.containsKey(column)) {
			String format = numberFormats.get(column);
			return setNumberFormatting(value, format, getFormatLocale());
		} // date/time format
		else if (dateTimeFormats.containsKey(column)) {
			Object formatFrom = dateTimeFormats.get(column)[0];
			Object formatTo = dateTimeFormats.get(column)[1];
			boolean isStandardFormat = (Boolean) dateTimeFormats.get(column)[2];

			if (isStandardFormat) {
				return setStandardDateTimeFormatting(value, formatFrom,
						formatTo, getFormatLocale());
			} else {
				return setCustomDateTimeFormatting(value, formatFrom, formatTo,
						getFormatLocale());
			}
		}

		return value;
	}

	private void checkVisibleColumns() {
		if (columnWidths != null && !columnWidths.isEmpty()) {
			columnVisibles.clear();
			firstColumn = enableAllColumns ? 0 : firstColumn;

			int width = 0;
			tableWidth += columnWidths.get(firstColumn);
			for (int column = firstColumn; column < allColumns; column++) {
				if (!isColumnHidden(column) && !isColumnFrozen(column)) {
					if (!enableAllColumns) {
						columnVisibles.add(column);
						if ((width += columnWidths.get(column)) > tableWidth) {
							break;
						}
					} else {
						columnVisibles.add(column);
					}
				}
			}

			if (!enableAllColumns && width < tableWidth) {
				tableWidth -= columnWidths.get(firstColumn);
				for (int column = firstColumn - 1; 0 <= column; column--) {
					if (!isColumnHidden(column) && !isColumnFrozen(column)) {
						columnVisibles.add(column);
						width += columnWidths.get(column);
						left -= columnWidths.get(column);
						if (width > tableWidth) {
							firstColumn = column;
							break;
						}
					}
				}
			}
			Collections.sort(columnVisibles);
		}
	}

	private Integer[] getVisibleColumns() {
		return columnVisibles.toArray(new Integer[] {});
	}

	private void calculateLeftAndTopValues() {
		int width = 0;
		for (int column = 0; column < allColumns; column++) {
			if (!isColumnHidden(column) && !isColumnFrozen(column)) {
				if ((width + columnWidths.get(column)) > left) {
					firstColumn = column;
					left = width;
					break;
				}
				width += columnWidths.get(column);
			}
		}

		firstRow = top / bodyRowHeight;
		top = firstRow * bodyRowHeight;
	}

	public void setPageSize(int size) {
		pageSize = size > 0 ? size : DEFAULT_PAGE_SIZE;
		paintContentObjects.put("pageSize", pageSize);
	}

	public void setPagerPages(int pages) {
		pagerPages = pages > 0 ? pages : DEFAULT_PAGER_PAGES;
		paintContentObjects.put("pagerTrayPages", pagerPages);
	}

	public void setFrozenTableMinimalWidth(int width) {
		tableFrozenMinimalWidth = width > 0 ? width
				: DEFAULT_TABLE_FROZEN_MINIMAL_WIDTH;
		paintContentObjects.put("tableFrozenMinimalWidth",
				tableFrozenMinimalWidth);
	}

	public void setTableMinimalWidth(int width) {
		tableMinimalWidth = width > 0 ? width : DEFAULT_TABLE_MINIMAL_WIDTH;
		paintContentObjects.put("tableMinimalWidth", tableMinimalWidth);
	}

	public void setTableHeight(int height) {
		// table height minus 2 (2 = borders)
		tableHeight = height > 0 ? (height - 2) : DEFAULT_TABLE_HEIGHT;
		paintContentObjects.put("tableHeight", tableHeight);
	}

	public void setFixedTableHeightEnabled(boolean enable) {
		paintContentObjects.put("enableFixedTableHeight", enable);
	}

	public void setHeaderRowHeight(int height) {
		headerRowHeight = height > 0 ? height : DEFAULT_HEADER_ROW_HEIGHT;
		paintContentObjects.put("headerRowHeight", headerRowHeight);
	}

	public void setBodyRowHeight(int height) {
		bodyRowHeight = height > 0 ? height : DEFAULT_HEADER_ROW_HEIGHT;
		paintContentObjects.put("bodyRowHeight", bodyRowHeight);
	}

	public void setColumnDefaultWidth(int width) {
		columnDefaultWidth = width > 0 ? width : DEFAULT_COLUMN_WIDTH;
	}

	private void checkColumnWidths() {
		if (columnWidths.size() < allColumns) {
			for (int column = 0; column < allColumns; column++) {
				if (!columnWidths.containsKey(column)) {
					columnWidths.put(column, columnDefaultWidth);
				}
			}
		}
	}

	private Integer[] getColumnWidths() {
		List<Integer> widths = new ArrayList<Integer>();
		for (int column = 0; column < allColumns; column++) {
			widths.add(columnWidths.get(column));
		}
		return widths.toArray(new Integer[] {});
	}

	public void setColumnWidths(int... widths) {
		for (int column = 0; column < widths.length; column++) {
			columnWidths.put(column, widths[column] > 0 ? widths[column]
					: columnDefaultWidth);
		}
	}

	public void setColumnSameWidths(int width, int... columns) {
		for (int column : columns) {
			if (column >= 0) {
				columnWidths
						.put(column, width > 0 ? width : columnDefaultWidth);
			}
		}
	}

	public void setColumnWidths(int column, int width) {
		if (column >= 0) {
			columnWidths.put(column, width > 0 ? width : columnDefaultWidth);
		}
	}

	private String[] getColumnHeaders() {
		return columnHeaders.toArray(new String[] {});
	}

	public void setColumnHeaders(String... headers) {
		if (headers != null && headers.length > 0) {
			for (String header : headers) {
				columnHeaders.add(header);
			}

			allColumns = columnHeaders.size();
			paintContentObjects.put("columnHeaders", getColumnHeaders());
		}
	}

	private void checkColumnHeadersId() {
		if (columnHeadersId == null
				|| columnHeadersId.size() < columnHeaders.size()) {
			for (int column = columnHeadersId.size(); column < columnHeaders
					.size(); column++) {
				columnHeadersId.add(columnHeaders.get(column));
			}
		}
	}

	public void setColumnHeadersId(String... headersId) {
		if (headersId != null && headersId.length > 0) {
			for (String headerId : headersId) {
				columnHeadersId.add(headerId);
			}
		}
	}

	private Integer[] getHeaderAlignments() {
		if (headerAlignments.size() < allColumns) {
			for (int column = 0; column < allColumns; column++) {
				if (!headerAlignments.containsKey(column)) {
					headerAlignments.put(column,
							headerDefaultAlignment.getAlignment());
				}
			}
		}

		List<Integer> alignments = new ArrayList<Integer>();
		for (int column = 0; column < allColumns; column++) {
			alignments.add(headerAlignments.get(column));
		}
		return alignments.toArray(new Integer[] {});
	}

	public void setHeaderDefaultAlignment(HorizontalAlignment alignment) {
		if (alignment != null) {
			headerDefaultAlignment = alignment;
		}
	}

	public void setHeaderAlignments(HorizontalAlignment... alignments) {
		for (Integer column = 0; column < alignments.length; column++) {
			headerAlignments
					.put(column,
							alignments[column] == null ? headerDefaultAlignment
									.getAlignment() : alignments[column]
									.getAlignment());
		}
	}

	public void setHeaderSameAlignments(HorizontalAlignment alignment,
			int... columns) {
		for (int column : columns) {
			if (alignment != null && column >= 0) {
				headerAlignments.put(column, alignment.getAlignment());
			}
		}
	}

	public void setHeaderAlignment(int column, HorizontalAlignment alignment) {
		if (alignment != null && column >= 0) {
			headerAlignments.put(column, alignment.getAlignment());
		}
	}

	public void setColumnDefaultAlignment(HorizontalAlignment alignment) {
		if (alignment != null) {
			columnDefaultAlignment = alignment;
		}
	}

	private Integer[] getColumnAlignments() {
		if (columnAlignments.size() < allColumns) {
			for (int column = 0; column < allColumns; column++) {
				if (!columnAlignments.containsKey(column)) {
					columnAlignments.put(column,
							columnDefaultAlignment.getAlignment());
				}
			}
		}

		List<Integer> alignments = new ArrayList<Integer>();
		for (int column = 0; column < allColumns; column++) {
			alignments.add(columnAlignments.get(column));
		}
		return alignments.toArray(new Integer[] {});
	}

	public void setColumnAlignments(HorizontalAlignment... alignments) {
		for (int column = 0; column < alignments.length; column++) {
			if (alignments[column] != null) {
				columnAlignments.put(column, alignments[column].getAlignment());
			}
		}
	}

	public void setColumnSameAlignments(HorizontalAlignment alignment,
			int... columns) {
		for (int column : columns) {
			if (alignment != null && column >= 0) {
				columnAlignments.put(column, alignment.getAlignment());
			}
		}
	}

	public void setColumnAlignment(int column, HorizontalAlignment alignment) {
		if (alignment != null && column >= 0) {
			columnAlignments.put(column, alignment.getAlignment());
		}
	}

	private boolean isColumnFrozen(int column) {
		if (!enableFrozenTable && columnFrozens != null) {
			columnFrozens.clear();
		}
		return columnFrozens.contains(column);
	}

	private Integer[] getFrozenColumns() {
		if (columnFrozens != null && !columnFrozens.isEmpty()) {
			if (!enableFrozenTable) {
				columnFrozens.clear();
				return null;
			}

			List<Integer> columns = new ArrayList<Integer>();
			for (int column : columnHiddens) {
				if (columnFrozens.contains(column)) {
					columnFrozens.remove(columnFrozens.indexOf(column));
				}
			}

			for (int column : columnFrozens) {
				if (column < allColumns) {
					columns.add(column);
				}
			}
			Collections.sort(columns);
			return columns.toArray(new Integer[] {});
		} else {
			return null;
		}
	}

	public void setColumnFrozens(int... columns) {
		for (int column : columns) {
			if (column >= 0 && !columnFrozens.contains(column)) {
				columnFrozens.add(column);
			}
		}
	}

	private boolean isColumnHidden(int column) {
		return columnHiddens.contains(column);
	}

	private Integer[] getColumnHiddens() {
		if (columnHiddens != null && !columnHiddens.isEmpty()) {
			for (int column : columnHiddens) {
				if (isColumnFrozen(column)) {
					columnFrozens.remove(columnFrozens.indexOf(column));
				}
			}

			List<Integer> columns = new ArrayList<Integer>();
			for (int column : columnHiddens) {
				if (column < allColumns) {
					columns.add(column);
				}
			}

			Collections.sort(columns);
			return columns.toArray(new Integer[] {});
		} else {
			return null;
		}
	}

	public void setColumnHiddens(int... columns) {
		for (int column : columns) {
			if (column >= 0 && !columnHiddens.contains(column)) {
				columnHiddens.add(column);
			}
		}
	}

	private Integer[] getColumnFormatted() {
		if (columnFormatted != null && !columnFormatted.isEmpty()) {
			List<Integer> columns = new ArrayList<Integer>();
			for (int column : columnFormatted) {
				if (column < allColumns) {
					columns.add(column);
				}
			}

			Collections.sort(columns);
			return columns.toArray(new Integer[] {});
		} else {
			return null;
		}
	}

	private void setColumnFormats(int... columns) {
		for (int column : columns) {
			if (column >= 0 && !columnFormatted.contains(column)) {
				columnFormatted.add(column);
			}
		}
	}

	private Locale getFormatLocale() {
		if (formattingLocale == null) {
			WebApplicationContext context = ((WebApplicationContext) getParent()
					.getApplication().getContext());
			WebBrowser webBrowser = context.getBrowser();
			formattingLocale = webBrowser.getLocale();
		}
		formattingLocale = formattingLocale == null ? Locale.ENGLISH
				: formattingLocale;
		return formattingLocale;
	}

	public void setFormatingLocale(String language) {
		Locale oldLocale = formattingLocale;
		try {
			formattingLocale = new Locale(language);
		} catch (Exception exception) {
			formattingLocale = oldLocale;
		}
	}

	public void setFormattingLocale(String language, String country) {
		Locale oldLocale = formattingLocale;
		try {
			formattingLocale = new Locale(language, country);
		} catch (Exception exception) {
			formattingLocale = oldLocale;
		}
	}

	private String setCustomDateTimeFormatting(String dateTime,
			Object formatFrom, Object formatTo, Locale locale) {
		String origDateTime = new String(dateTime);
		try {
			SimpleDateFormat formattingFrom = new SimpleDateFormat(
					formatFrom.toString());
			SimpleDateFormat formattingTo = new SimpleDateFormat(
					formatTo.toString(), locale);
			Date date = formattingFrom.parse(dateTime);
			return formattingTo.format(date);
		} catch (Exception exception) {
			return origDateTime;
		}
	}

	private String setStandardDateTimeFormatting(String dateTime,
			Object formatFrom, Object formatTo, Locale locale) {
		String origDateTime = new String(dateTime);
		try {
			int formattingInstance = (Integer) formatTo / 4;
			int formattingType = (Integer) formatTo % 4;
			SimpleDateFormat formattingFrom = new SimpleDateFormat(
					(String) formatFrom);
			DateFormat formattingTo = null;
			Date date = formattingFrom.parse(dateTime);
			switch (formattingInstance) {
			case 0:
				formattingTo = DateFormat.getTimeInstance(formattingType,
						locale);
				break;
			case 1:
				formattingTo = DateFormat.getDateInstance(formattingType,
						locale);
				break;
			case 2:
				formattingTo = DateFormat.getDateTimeInstance(formattingType,
						formattingType, locale);
				break;
			default:
				break;
			}
			return formattingTo.format(date);
		} catch (Exception exception) {
			return origDateTime;
		}
	}

	public void setDateTimeFormats(String formatFrom, String formatTo,
			int... columns) {
		for (int column : columns) {
			if (formatFrom != null && !formatFrom.isEmpty() && formatTo != null
					&& !formatTo.isEmpty() && column >= 0) {
				dateTimeFormats.put(column, new Object[] { formatFrom,
						formatTo, Boolean.FALSE });
			}
		}
		setColumnFormats(columns);
	}

	public void setStandardDateTimeFormats(String formatFrom,
			DateTimeFormat formatTo, int... columns) {
		for (int column : columns) {
			if (formatFrom != null && !formatFrom.isEmpty() && formatTo != null
					&& !formatTo.equals(DateTimeFormat.UNDEFINED_FORMAT)
					&& column >= 0) {
				dateTimeFormats.put(column,
						new Object[] { formatFrom, formatTo.getFromat(),
								Boolean.TRUE });
			}
		}
		setColumnFormats(columns);
	}

	private String setNumberFormatting(String number, String format,
			Locale locale) {
		String origNumber = number;
		try {
			DecimalFormat formatting = (DecimalFormat) NumberFormat
					.getNumberInstance(getLocale());
			formatting.applyPattern(format);
			return formatting.format(Float.parseFloat(number));
		} catch (Exception exception) {
			return origNumber;
		}
	}

	private String[] getColumnSortingStrings() {
		if (columnSortings != null && !columnSortings.isEmpty()) {
			List<String> columns = new ArrayList<String>();
			for (String column : columnSortings) {
				if (Math.abs(Integer.parseInt(column.substring(1))) >= 0
						&& Math.abs(Integer.parseInt(column.substring(1))) < allColumns) {
					columns.add(column);
				}
			}
			return columns.toArray(new String[] {});
		} else {
			return null;
		}
	}

	private Sorting[] getColumnSortings() {
		if (columnSortings != null && !columnSortings.isEmpty()) {
			List<Integer> sortingIndex = new ArrayList<Integer>();
			for (String column : columnSortings) {
				sortingIndex.add(Integer.parseInt(column.substring(1)));
			}
			Collections.sort(sortingIndex);

			Sorting sorting = null;
			List<Sorting> sortings = new ArrayList<Sorting>();
			List<Sorting> frozenSortings = new ArrayList<Sorting>();
			for (Integer index : sortingIndex) {
				for (String column : columnSortings) {
					sorting = new Sorting(column, columnHeaders.get(index),
							columnHeadersId.get(index));
					if (sorting.getColumn().equals(index)) {
						break;
					}
				}
				if (isColumnFrozen(index)) {
					frozenSortings.add(sorting);
				} else {
					sortings.add(sorting);
				}
			}

			List<Sorting> finalSortings = new ArrayList<Sorting>();
			if (!frozenSortings.isEmpty()) {
				finalSortings.addAll(frozenSortings);
			}
			if (!sortings.isEmpty()) {
				finalSortings.addAll(sortings);
			}

			return finalSortings.toArray(new Sorting[] {});
		} else {
			return null;
		}
	}

	public void setColumnSortings(String type, int... columns) {
		if (type.equals(Sorting.ASCEDING) || type.equals(Sorting.DESCEDING)) {
			for (int column : columns) {
				if (column >= 0) {
					if (columnSortings.contains(Sorting.ASCEDING + column)) {
						columnSortings.remove(Sorting.ASCEDING + column);
					} else if (columnSortings.contains(Sorting.DESCEDING
							+ column)) {
						columnSortings.remove(Sorting.DESCEDING + column);
					}
					columnSortings.add(type + column);
				}
			}
		}
	}

	public String setImage(String url, int width, int height) {
		return ("image_:_" + url + "_:_" + width + "_:_" + height);
	}

	public void setNumberFormats(String format, int... columns) {
		for (int column : columns) {
			if (format != null && !format.isEmpty() && column >= 0) {
				numberFormats.put(column, format);
			}
		}
		setColumnFormats(columns);
	}

	public void setLazyLoadingEnabled(boolean enable) {
		paintContentObjects.put("enableAllRows", enableAllRows = !enable);
	}

	public void setAllRowsEnabled(boolean enable) {
		paintContentObjects.put("enableAllRows", enableAllRows = enable);
	}

	public void setAllColumnsEnabled(boolean enable) {
		paintContentObjects.put("enableAllColumns", enableAllColumns = enable);
	}

	public void setColumnNumberInWaitBoxEnabled(boolean enable) {
		paintContentObjects.put("enableColumnNumberInWaitBox", enable);
	}

	public void setExpandedFrozenHeaderCellEnabled(boolean enable) {
		paintContentObjects.put("enableExpandedFrozenHeaderCell", enable);
	}

	public void setExpandedFrozenHeaderCellForActiveColumnEnabled(boolean enable) {
		paintContentObjects.put(
				"enableExpandedFrozenHeaderCellForActiveColumn", enable);
	}

	public void setExpandedFrozenBodyCellEnabled(boolean enable) {
		paintContentObjects.put("enableExpandedFrozenBodyCell", enable);
	}

	public void setExpandedHeaderCellEnabled(boolean enable) {
		paintContentObjects.put("enableExpandedHeaderCell", enable);
	}

	public void setExpandedHeaderCellForActiveColumnEnabled(boolean enable) {
		paintContentObjects.put("enableExpandedHeaderCellForActiveColumn",
				enable);
	}

	public void setExpandedBodyCellEnabled(boolean enable) {
		paintContentObjects.put("enableExpandedBodyCell", enable);
	}

	public void setWordWrapInHeaderCellsEnabled(boolean enable) {
		// TODO enable
		paintContentObjects.put("enableWordWrapInHeaderCells", false);
	}

	public void setWordWrapInBodyCellsEnaled(boolean enable) {
		// TODO enable
		paintContentObjects.put("enableWordWrapInBodyCells", false);
	}

	public void setFrozenTableEnabled(boolean enable) {
		paintContentObjects
				.put("enableFrozenTable", enableFrozenTable = enable);
	}

	public void setHideFrozenTableEnabled(boolean enable) {
		paintContentObjects.put("enableHideFrozenTable", enable);
	}

	public void setFilteredFrozenTableEnabled(boolean enable) {
		paintContentObjects.put("enableFilteredFrozenTable", enable);
	}

	public void setFilteredTableEnabled(boolean enable) {
		paintContentObjects.put("enableFilteredTable", enable);
	}

	public void setEmptyFrozenTableEnabled(boolean enable) {
		paintContentObjects.put("enableEmptyFrozenTable", enable);
	}

	public void setResizeFrozenTableEnabled(boolean enable) {
		paintContentObjects.put("enableResizeFrozenTable", enable);
	}

	public void setSortedFrozenTableEnabled(boolean enable) {
		paintContentObjects.put("enableSortedFrozenTable", enable);
	}

	public void setSortedTableEnabled(boolean enable) {
		paintContentObjects.put("enableSortedTable", enable);
	}

	public void setFilterHelperEnabled(boolean show) {
		paintContentObjects.put("enableFilterHelper", show);
	}

	public void setFilterHelperOnTopEnabled(boolean show) {
		paintContentObjects.put("enableFilterHelperOnTop", show);
	}

	public void setPageableTableEnabled(boolean enable) {
		paintContentObjects.put("enablePageableTable", enable);
	}

	public void setScrollableTableEnabled(boolean enable) {
		paintContentObjects.put("enablePageableTable", !enable);
	}

	public void setFreezeAnyColumnsEnabled(boolean enable) {
		paintContentObjects.put("enableFreezeAnyColumns", enable);
	}

	public void setUnfreezeAnyColumnsEnabled(boolean enable) {
		paintContentObjects.put("enableUnfreezeAnyColumns", enable);
	}

	public void setHorizontalLinesEnabled(boolean enable) {
		paintContentObjects.put("enableHorizontalLines", enable);
	}

	public void setVerticalLinesEnabled(boolean enable) {
		paintContentObjects.put("enableVerticalLines", enable);
	}

	public void setFreezeColumnsEnabled(boolean enable, int... columns) {
		int enableFreeze = enable ? 1 : -1;
		for (int column : columns) {
			if (columnFreezeEnabled.contains(column)) {
				columnFreezeEnabled.remove(columnFreezeEnabled.indexOf(column));
			} else if (columnFreezeEnabled.contains(-column)) {
				columnFreezeEnabled
						.remove(columnFreezeEnabled.indexOf(-column));
			}
			columnFreezeEnabled.add(enableFreeze * column);
		}
	}

	public void setUnfreezeColumnsEnabled(boolean enable, int... columns) {
		int enableFreeze = enable ? 1 : -1;
		for (int column : columns) {
			if (columnUnfreezeEnabled.contains(column)) {
				columnUnfreezeEnabled.remove(columnUnfreezeEnabled
						.indexOf(column));
			} else if (columnUnfreezeEnabled.contains(-column)) {
				columnUnfreezeEnabled.remove(columnUnfreezeEnabled
						.indexOf(-column));
			}
			columnUnfreezeEnabled.add(enableFreeze * column);
		}
	}

	public void setHighlightActiveRowEnabled(boolean enable) {
		paintContentObjects.put("enableHighlightActiveRow", enable);
	}

	public void setLoadingMessage(String message) {
		if (message != null && !message.isEmpty()) {
			paintContentObjects.put("loadingMessage", message);
		}
	}

	public void setLoadingMessageOpacity(float opacity) {
		if (0 <= opacity && opacity <= 1) {
			paintContentObjects.put("loadingMessageOpacity", opacity);
		}
	}

	public void setLoadingMessageEnabled(boolean enable) {
		paintContentObjects.put("enableLoadingMessage", enable);
	}

	public void setLoadingMessageDelay(int showDelay, int hideDelay) {
		if (showDelay > 0 && hideDelay > 0) {
			paintContentObjects.put("loadingMessageShowDelay", showDelay);
			paintContentObjects.put("loadingMessageHideDelay", hideDelay);
		}
	}

	public void setEmptyDataMessage(String message) {
		if (message != null && !message.isEmpty()) {
			paintContentObjects.put("emptyDataMessage", message);
		}
	}

	public void setNoDataMessage(String noDataTitle, String noDataMessage) {
		if (noDataTitle != null && !noDataTitle.isEmpty()
				&& noDataMessage != null && !noDataMessage.isEmpty()) {
			this.noDataMessageTitle = noDataTitle;
			this.noDataMessage = noDataMessage;
		}
	}

	public void setErrorMessage(String errorTitle, String errorMessage) {
		if (errorTitle != null && !errorTitle.isEmpty() && errorMessage != null
				&& !errorMessage.isEmpty()) {
			this.errorMessageTitle = errorTitle;
			this.errorMessage = errorMessage;
		}
	}

}
